import pluralize from 'pluralize';
import { 
    ActionContract,
    ToolbarContract,
    XformToFormContract,
    XformToTableContract,
    XformToDocketContract,
    FormContract,
    TableContract,
    DocketContract
} from '../contracts';
import { 
    Agent,
    CleverQueryService } from "../services";
import { FormFactory } from "./form-factory";

export abstract class Model<M> 
    extends Agent 
        implements 
            XformToFormContract, 
            XformToTableContract,
            XformToDocketContract{
                
    creatable: boolean = true;
    soloable: boolean = true;
    collectionable: boolean = true;
    deletable: boolean = true;
    editable: boolean = true;
    statusChangeable: boolean = true;
    soloType: string | 'docket' = 'docket';
    collectionType: string | 'table' | 'lists' | 'cards' = 'lists';

    modelName: string;
    model: M;
    models: M[] = [];
    protected actions: ActionContract[];
    protected toolbars: ToolbarContract[];
    protected exclude: string[] = ['status', 'creatorId', 'createdAt', 'updatedAt'];
    
    protected abstract transform(datum: Object): M;

    toModel(datum: Object){
        this.model = this.transform(datum);
        return this;
    }

    toModels(data: Object[]){
        this.models = [];
        data.forEach((datum) => {
            this.models.push(this.transform(datum));
        });
        return this;
    }

    setActions(actions: ActionContract[]){
        this.actions = (this.store.get('authType') == 'admin') ? actions : [];
        return this;
    }
    
    setToolbars(toolbars: ToolbarContract[]){
        this.toolbars =  (this.store.get('authType') == 'admin') ? toolbars : [];
        return this;
    }
    
    toForm(columns: Object[], resource: Object = {}): FormContract{
        return (new FormFactory(columns, resource, this.exclude)).make();
    }

    toTable(): TableContract{
        let rows: string[][] = [],
            columns: string[],
            tmpRow: string[];
        this.models.forEach((model) => {
            tmpRow = [];
            if(!columns){
                columns = Object.keys(model).filter((column)=>{
                    return !this.exclude.includes(column);
                });
            }
            for(let column in model){
                if(!this.exclude.includes(column)){
                    tmpRow.push(model[<string> column]);
                }
            }
            rows.push(tmpRow);
        });
        return {
            title: null,
            columns: columns,
            rows: rows,
            summary: null,
            image: { title: null, url: null },
            meta: null,
            actions: this.actions,
            toolbars: this.toolbars
        };
    }
    
    toDocket(): DocketContract{
        let rows: string[][] = [];
        for(let cell in this.model){
            if(!this.exclude.includes(<string> cell)){
                rows.push([<string> cell, <string> this.model[ <string> cell]]);
            }
        }
        return {
            title: null,
            columns: ['Key', 'Value'],
            rows: rows,
            summary: null,
            image: { title: null, url: null },
            meta: null,
            actions: this.actions,
            toolbars: this.toolbars
        };
    }

    one(id: number, callback: (response) => void, actions: ActionContract[] = [], toolbars: ToolbarContract[] = []){
        this.go((agent) => {
            agent.resource(pluralize(this.modelName), id).row((response) => {
                //xform json response to model, then to the solo type set on the model 
                let solo = this.setActions(actions).setToolbars(toolbars).toModel(response[this.modelName])['to' + this.soloType.charAt(0).toUpperCase() + this.soloType.slice(1)]();
                callback(solo);
            }); 
        }); 
    }

    many(query: any,  callback: (collection, total) => void, actions: ActionContract[] = [], toolbars: ToolbarContract[] = []){
        this.go((agent) => {
            agent.resource(pluralize(this.modelName)).rows((response) => {
                //xform json response to model, then to the collection type set on the model 
                let collection = this.setActions(actions).setToolbars(toolbars).toModels(response[pluralize(this.modelName)])['to' + this.collectionType.charAt(0).toUpperCase() + this.collectionType.slice(1)]();
                callback(collection, response['subTotal']);
            }, query); 
        }); 
    }

    relatives(query: any, id: number, model: string, callback: (collection, total) => void, actions: ActionContract[] = [], toolbars: ToolbarContract[] = []){
        // this.modelName = relative name
        this.go((agent) => {
            agent.aGet(`${agent.baseUrl}/${pluralize(model)}/${id}/${pluralize(this.modelName)}`, (response) => {
                //xform json response to model, then to the collection type set on the model 
                let collection = this.setActions(actions).setToolbars(toolbars).toModels(response[pluralize(this.modelName)])['to' + this.collectionType.charAt(0).toUpperCase() + this.collectionType.slice(1)]();
                callback(collection, response['subTotal']);
            }, query); 
        });
    }

    form(callback: (response) => void, resource: Object = {}){
        this.go((agent) => {
            agent.resource(pluralize(this.modelName)).columns((response) => {
                let formData = this.toForm(response[this.modelName], resource);
                callback(formData);
            });
        }); 
    }
}