import { Injectable } from '@angular/core';
import { 
  Http, 
  XHRBackend, 
  Headers, 
  Response, 
  RequestOptions, 
  RequestOptionsArgs, 
  Request } from '@angular/http';

import 'rxjs/add/operator/map';
import 'rxjs/add/observable/throw';
import { Observable } from 'rxjs/Observable';

import { Resource, Person } from './models';
import { BagContract } from '../../contracts';

import { ToastyWrapperService } from '../toasty-wrapper';
import { SlimLoadingBarService } from 'ng2-slim-loading-bar';
import { StorageService } from '../storage';
import { CleverQueryService } from '../clever-query';
import { SB_CORE_CONFIG, BASE_URL } from '../../sb-core.config';

@Injectable()
export class Agent extends Http{
  
  config: any;
  urls: any;
  headers: Headers;
  baseDomain: string;
  baseOptions: RequestOptionsArgs;
  accessToken: string = null;
  baseUrl: string = BASE_URL;

  constructor(private toastMaker: ToastyWrapperService,
              private slimLoader: SlimLoadingBarService,
              protected store: StorageService,
              backend: XHRBackend,
              options: RequestOptions,
              protected cleverQuery: CleverQueryService
            ){
    super(backend, options);
    this.store = this.store.local();

    //initialize request options headers using values from config
    let configHeaders = SB_CORE_CONFIG['headers'];
    let baseHeaders = new Headers();
    configHeaders.forEach(function(header){
      baseHeaders.append(header.key, header.value);
    });

    //add jwt token to header
    baseHeaders.append('Authorization', `Bearer ${this.store.get('authToken')}`);

    //make header ready for every request
    this.headers = baseHeaders;
    this.baseOptions = {
      headers : baseHeaders
    };
    
    //set configs
    this.config = SB_CORE_CONFIG;
    this.urls = SB_CORE_CONFIG['urls'];
    this.baseDomain = BASE_URL;

  }
  
  /*
   * high-level features 
   */
  go(callback: (agent: Agent) => void){
    callback(this);
    
    // if(!this.accessToken){
    //   this.aPost(this.urls.client.token, {
    //       security_key: this.config.agent.key,
    //       security_token: this.config.agent.token
    //   }, (response) => {
    //     this.accessToken = response.token;
  
    //     // request to server was succesfully
    //     if(!!this.accessToken){
    //       callback(this);
    //     }
    
    //     // something went wrong with request to server
    //     else{
    //       console.log('api error');
    //     }
    //   });  
    // }
    // else{
    //   // request to server was succesfully
    //   if(!!this.accessToken){
    //     callback(this);
    //   }
  
    //   // something went wrong with request to server
    //   else{
    //     console.log('api error');

    //   }
    // }
  }

  version(callback: (response) => void){
    this.aGet(this.urls.version, (response) => {
        callback(response);
    });  
  } 

  /*
   * model features 
   */

  person(type: string, id: number = null){
    return new Person(this, type, id);
  }

  resource(type: string, id: number = null){
    return new Resource(this, type, id);
  }

  /*
   * base features  
   */

  aRequest(url: string | Request, callback: (response, url) => any, moreOptions?: RequestOptionsArgs){
    this.onStart();
    return this.request(
        url, this.unifyOptions(this.baseOptions, moreOptions)
      )
      .map(this.jsonify)
      .subscribe((response) => {
        callback(response, url)
      }, this.onError.bind(this), this.onFinish.bind(this));
  }

  aGet(url: string, callback: (response, url) => any, moreOptions?: RequestOptionsArgs){
    this.onStart();
    return this.get(
        url, this.unifyOptions(this.baseOptions, moreOptions)
      )
      .map(this.jsonify)
      // .retry(2)
      // .timeout(60*1000)
      .subscribe((response) => {
        callback(response, url)
      }, this.onError.bind(this), this.onFinish.bind(this));
  }

  aPost(url: string, payload: Object, callback: (response, url) => any, moreOptions?: RequestOptionsArgs){
    this.onStart();
    return this.post(
        url, `payload=${JSON.stringify(payload)}`, this.unifyOptions(this.baseOptions, moreOptions)
      )
      .map(this.jsonify)
      .subscribe((response) => {
        callback(response, url)
      }, this.onError.bind(this), this.onFinish.bind(this));
  }

  aPut(url: string, payload: Object, callback: (response, url) => any, moreOptions?: RequestOptionsArgs){
    this.onStart();
    return this.put(
        url, `payload=${JSON.stringify(payload)}`, this.unifyOptions(this.baseOptions, moreOptions)
      )
      .map(this.jsonify)
      .subscribe((response) => {
        callback(response, url)
      }, this.onError.bind(this), this.onFinish.bind(this));
  }
  
  aDelete(url: string, callback: (response, url) => any, moreOptions?: RequestOptionsArgs){
    this.onStart();
    return this.delete(
        url, this.unifyOptions(this.baseOptions, moreOptions)
      )
      .map(this.jsonify)
      .subscribe((response) => {
        callback(response, url)
      }, this.onError.bind(this), this.onFinish.bind(this));
  }
  
  aPatch(url: string, payload: Object, callback: (response, url) => any, moreOptions?: RequestOptionsArgs){
    this.onStart();
    return this.patch(
        url, `payload=${JSON.stringify(payload)}`, this.unifyOptions(this.baseOptions, moreOptions)
      )
      .map(this.jsonify)
      .subscribe((response) => {
        callback(response, url)
      }, this.onError.bind(this), this.onFinish.bind(this));
  }

  /*
   * helper features
   */
  
  protected unifyOptions(base:RequestOptionsArgs = null, more: RequestOptionsArgs = null){
    // return this._requestOptions.merge(options);
    return base;
  }
  
  //jsonify Http Response
  protected jsonify(response: Response) {
    return response.json() || Object;
  }

  protected addAuthToken(url: string): string{
    let authToken = this.store.get('authToken');
    return (url.includes('?')) ? `${url}&auth_token=${authToken}` : `${url}?auth_token=${authToken}`;
  }

  protected urlify(base: string, model: string = null, id: number = null, action: string = 'id', slugs: BagContract[] = null): string{
    //start with base domain
    let url: string = base;
    
    //add model slug
    if(model){
      url += '/' + model;
    }

    //add model id
    if(id){
      url += '/' + id;
    }

    //add action
    if(action){
      url += '/' + action;
    }

    //add slugs
    if(slugs.length > 0){
      let slugStr = '?';
      slugs.forEach(function(slug, index, slugs){
        slugStr += `${slug.key}=${slug.value}`;
        if(index != slugs.length){
          slugStr += '&';
        }
      });
      
      url += slugStr;
    }
    return url;
  }    

  protected onError (error: Response | any){
    let errMsg: string;
    if (error instanceof Response) {
        const body = error.json() || '';
        const err = body.error || body;
        // errMsg = `${error.status} - ${error.statusText || ''} ${err}`;
        errMsg = `${error.status} - ${error.statusText || ''}`;
    } else {
        errMsg = error.message ? error.message : error.toString();
    }
      
    this.toastMaker.toast('HTTP Error', errMsg);
    this.slimLoader.complete();
    
    return Observable.throw(errMsg);
  }
  
  protected onStart(){
      this.slimLoader.start(() => {
      });
      this.toastMaker.toast('HTTP Action', 'Connecting...');
  }

  protected onFinish(){
    this.slimLoader.complete();
    this.toastMaker.toast('HTTP Action', 'Connecting...Finished.');
  }
}